/* object based disk file system
 * 
 * This software is licensed under the GPL.  See the file COPYING in the
 * top directory of this distribution for details.
 * 
 * Copyright (C), 1999, Stelias Computing Inc
 *
 *
 */


#ifndef _OBDFS_H
#define OBDFS_H
#include <linux/obd_class.h>
#include <linux/list.h>

struct obdfs_pgrq {
	struct list_head	 rq_plist;	/* linked list of req's */
	unsigned long            rq_jiffies;
	struct page 		*rq_page;	/* page to be written */
};

struct list_head obdfs_super_list;	 /* list of all OBDFS superblocks */

struct obdfs_sb_info {
	struct list_head	 osi_list;	/* list of supers */
	struct obd_conn		 osi_conn;
	struct super_block	*osi_super;
	struct obd_device	*osi_obd;
	struct obd_ops		*osi_ops;
	ino_t			 osi_rootino;	/* number of root inode */
	int			 osi_minor;	/* minor of /dev/obdX */
	struct list_head	 osi_inodes;	/* list of dirty inodes */
	unsigned long		 osi_cache_count;
	struct semaphore         osi_list_mutex;
};

struct obdfs_inode_info {
	int		 oi_flags;
	struct list_head oi_inodes;
	struct list_head oi_pages;
	char 		 oi_inline[OBD_INLINESZ];
};

/* dir.c */
int obdfs_check_dir_entry (const char * function, struct inode * dir,
			  struct ext2_dir_entry_2 * de, struct page * page,
			  unsigned long offset);
extern struct file_operations obdfs_dir_operations;
extern struct inode_operations obdfs_dir_inode_operations;

/* file.c */
extern struct file_operations obdfs_file_operations;
extern struct inode_operations obdfs_file_inode_operations;

/* flush.c */
void obdfs_dequeue_pages(struct inode *inode);
int obdfs_flushd_init(void);
int obdfs_flushd_cleanup(void);
int obdfs_flush_reqs(struct list_head *inode_list, unsigned long check_time);
int obdfs_flush_dirty_pages(unsigned long check_time);

/* namei.c */
struct dentry *obdfs_lookup(struct inode * dir, struct dentry *dentry);
int obdfs_create (struct inode * dir, struct dentry * dentry, int mode);
int obdfs_mkdir(struct inode *dir, struct dentry *dentry, int mode);
int obdfs_rmdir(struct inode *dir, struct dentry *dentry);
int obdfs_unlink(struct inode *dir, struct dentry *dentry);
int obdfs_mknod(struct inode *dir, struct dentry *dentry, int mode, int rdev);
int obdfs_symlink(struct inode *dir, struct dentry *dentry,
		  const char *symname);
int obdfs_link(struct dentry *old_dentry, struct inode *dir,
	       struct dentry *dentry);
int obdfs_rename(struct inode *old_dir, struct dentry *old_dentry,
		 struct inode *new_dir, struct dentry *new_dentry);

/* rw.c */
int obdfs_do_writepage(struct inode *, struct page *, int sync);
int obdfs_init_pgrqcache(void);
void obdfs_cleanup_pgrqcache(void);
inline void obdfs_pgrq_del(struct obdfs_pgrq *pgrq);
int obdfs_readpage(struct dentry *dentry, struct page *page);
int obdfs_writepage(struct dentry *dentry, struct page *page);
struct page *obdfs_getpage(struct inode *inode, unsigned long offset,
			   int create, int locked);
int obdfs_write_one_page(struct file *file, struct page *page,
			 unsigned long offset, unsigned long bytes,
			 const char * buf);
int obdfs_do_vec_wr(struct inode **inodes, obd_count num_io, obd_count num_oa,
		    struct obdo **obdos, obd_count *oa_bufs,
		    struct page **pages, char **bufs, obd_size *counts,
		    obd_off *offsets, obd_flag *flags);
void obdfs_truncate(struct inode *inode);

/* super.c */
extern long obdfs_cache_count;

/* symlink.c */
extern struct inode_operations obdfs_fast_symlink_inode_operations;
extern struct inode_operations obdfs_symlink_inode_operations;

/* sysctl.c */
void obdfs_sysctl_init(void);
void obdfs_sysctl_clean(void);


static inline struct obdfs_inode_info *obdfs_i2info(struct inode *inode)
{
	return (struct obdfs_inode_info *)&(inode->u.generic_ip);
}

static inline struct obdfs_sb_info *obdfs_i2sbi(struct inode *inode)
{
	return (struct obdfs_sb_info *) &(inode->i_sb->u.generic_sbp);
}

static inline struct list_head *obdfs_iplist(struct inode *inode) 
{
	struct obdfs_inode_info *info = obdfs_i2info(inode);

	return &info->oi_pages;
}

static inline struct list_head *obdfs_islist(struct inode *inode) 
{
	struct obdfs_inode_info *info = obdfs_i2info(inode);

	return &info->oi_inodes;
}

static inline struct list_head *obdfs_slist(struct inode *inode) 
{
	struct obdfs_sb_info *sbi = obdfs_i2sbi(inode);

	return &sbi->osi_inodes;
}

#define obd_down(mutex) {\
	/* CDEBUG(D_INFO, "get lock\n"); */ \
	down(mutex);\
}

#define obd_up(mutex) {\
	up(mutex);\
	/* CDEBUG(D_INFO, "free lock\n"); */ \
}

/* We track if a page has been added to the OBD page cache by stting a
 * flag on the page.  We have chosen a bit that will hopefully not be
 * used for a while.
 */
#define PG_obdcache 29
#define OBDAddCachePage(page)	test_and_set_bit(PG_obdcache, &(page)->flags)
#define OBDClearCachePage(page)	clear_bit(PG_obdcache, &(page)->flags)

static inline void obdfs_print_plist(struct inode *inode) 
{
	struct list_head *page_list = obdfs_iplist(inode);
	struct list_head *tmp;

	CDEBUG(D_INFO, "inode %ld: page", inode->i_ino);
	/* obd_down(&obdfs_i2sbi(inode)->osi_list_mutex); */
	if (list_empty(page_list)) {
		CDEBUG(D_INFO, " list empty\n");
		obd_up(&obdfs_i2sbi(inode)->osi_list_mutex);
		return;
	}

	tmp = page_list;
	while ( (tmp = tmp->next) != page_list) {
		struct obdfs_pgrq *pgrq;
		pgrq = list_entry(tmp, struct obdfs_pgrq, rq_plist);
		CDEBUG(D_INFO, " %p", pgrq->rq_page);
	}
	CDEBUG(D_INFO, "\n");
	/* obd_up(&obdfs_i2sbi(inode)->osi_list_mutex); */
}

static inline int obdfs_has_inline(struct inode *inode)
{
	return (obdfs_i2info(inode)->oi_flags & OBD_FL_INLINEDATA);
}

static void inline obdfs_from_inode(struct obdo *oa, struct inode *inode)
{
	struct obdfs_inode_info *oinfo = obdfs_i2info(inode);

	CDEBUG(D_INFO, "src inode %ld, dst obdo %ld valid 0x%08x\n",
	       inode->i_ino, (long)oa->o_id, oa->o_valid);
	obdo_from_inode(oa, inode);
	if (obdfs_has_inline(inode)) {
		CDEBUG(D_INFO, "copying inline data from inode to obdo\n");
		memcpy(oa->o_inline, oinfo->oi_inline, OBD_INLINESZ);
		oa->o_obdflags |= OBD_FL_INLINEDATA;
		oa->o_valid |= OBD_MD_FLINLINE;
	}
} /* obdfs_from_inode */

static void inline obdfs_to_inode(struct inode *inode, struct obdo *oa)
{
	struct obdfs_inode_info *oinfo = obdfs_i2info(inode);

	CDEBUG(D_INFO, "src obdo %ld valid 0x%08x, dst inode %ld\n",
	       (long)oa->o_id, oa->o_valid, inode->i_ino);
	/* If the inode is dirty, we won't overwrite the data there, as it
	 * is newer than the data on the disk.  The ext2obd side only will
	 * change the block count, so we are guaranteed that is safe.
	 */
	if (inode->i_state & I_DIRTY) {
		CDEBUG(D_INODE, "dirty inode %ld, only copying blocks\n",
		       inode->i_ino);
		oa->o_valid = OBD_MD_FLBLOCKS;
	}

	obdo_to_inode(inode, oa);

	if (obdo_has_inline(oa)) {
		CDEBUG(D_INFO, "copying inline data from obdo to inode\n");
		memcpy(oinfo->oi_inline, oa->o_inline, OBD_INLINESZ);
		oinfo->oi_flags |= OBD_FL_INLINEDATA;
	}
} /* obdfs_to_inode */

#define NOLOCK 0
#define LOCKED 1

#ifdef OPS
#warning "*** WARNING redefining OPS"
#else
#define OPS(sb,op) ((struct obdfs_sb_info *)(& ## sb ## ->u.generic_sbp))->osi_ops->o_ ## op
#define IOPS(inode,op) ((struct obdfs_sb_info *)(& ## inode->i_sb ## ->u.generic_sbp))->osi_ops->o_ ## op
#endif

#ifdef ID
#warning "*** WARNING redefining ID"
#else
#define ID(sb) (&((struct obdfs_sb_info *)( & ## sb ## ->u.generic_sbp))->osi_conn)
#define IID(inode) (&((struct obdfs_sb_info *)( & ## inode->i_sb ## ->u.generic_sbp))->osi_conn)
#endif

#define OBDFS_SUPER_MAGIC 0x4711

#endif

